/*
 * z2-Environment
 * 
 * Copyright(c) ZFabrik Software GmbH & Co. KG
 * 
 * contact@zfabrik.de
 * 
 * http://www.z2-environment.eu
 */
package com.zfabrik.impl.components.java.jdt;
import java.io.File;
import java.io.InputStream;
import java.util.HashMap;
import java.util.Map;

import org.eclipse.jdt.internal.compiler.classfmt.ClassFileReader;
import org.eclipse.jdt.internal.compiler.classfmt.ClassFormatException;
import org.eclipse.jdt.internal.compiler.env.INameEnvironment;
import org.eclipse.jdt.internal.compiler.env.NameEnvironmentAnswer;

/**
 * input to the jdt compiler. A name env provides classes to link to, sources to compile...
 * @author hb
 */

public class NameEnvironmentImpl implements INameEnvironment {
	private ClassLoader cl;
	private File[] sourceFolders;
	private Map<String,File> sfnToFile = new HashMap<String, File>();
	private String encoding;

	public NameEnvironmentImpl(ClassLoader cl, String encoding, File... sourceFolders) {
		this.cl = cl;
		this.sourceFolders = sourceFolders;
		this.encoding = encoding;
	}

	public void cleanup() {
		this.sfnToFile.clear();
	}

	public NameEnvironmentAnswer findType(char[][] compoundName) {
		return _findType(concat(compoundName));
	}

	public NameEnvironmentAnswer findType(char[] typeName, char[][] packageName) {
		return _findType(new StringBuilder(200).append(concat(packageName)).append('/').append(typeName).toString());
	}

	public boolean isPackage(char[][] parentCompoundName, char[] packageName) {
		String slashedClassName = (parentCompoundName==null? 
			new String(packageName) :
			new StringBuilder().append(concat(parentCompoundName)).append('/').append(packageName).toString()
		);
		// check in our sources
		if (_findSourceFile(slashedClassName)!=null) {
			return false;
		}
		// is it a class?
		if (this.cl.getResource(slashedClassName+".class")!=null) {
			return false;
		}
		// if its last part starts with an upper case char, we assume it is no package
		if (packageName.length>0 && Character.isUpperCase(packageName[0])) {
			return false;
		}
		// no idea... better say yes. 
		return true;
	}

	// -------

	public static String concat(char[][] compoundName) {
		if (compoundName==null) {
			return "";
		}
		StringBuilder b = new StringBuilder(200);
		for (char[] s : compoundName) {
			if (b.length() > 0) {
				b.append('/');
			}
			b.append(s);
		}
		return b.toString();
	}

	private File _findFile(String name) {
		File f = this.sfnToFile.get(name);
		if (f!=null) {
			return f;
		}
		if (this.sourceFolders != null) {
			for (File sf : this.sourceFolders) {
				f = new File(sf, name);
				if (f.exists()  && f.getName().equals(name)) {
					this.sfnToFile.put(name,f);
					return f;
				}
			}
		}
		return null;
	}
	
	private File _findSourceFile(String slashedClassName) {
		return _findFile(slashedClassName+".java");
	}

	private NameEnvironmentAnswer _findType(String slashedClassName) {
		// try as source file first
		File sf = _findSourceFile(slashedClassName);
		if (sf != null) {
			return new NameEnvironmentAnswer(new FileCompilationUnit(sf,this.encoding), null);
		}
		// try as byte array next
		String rn = slashedClassName + ".class";
		InputStream in = this.cl.getResourceAsStream(rn);
		if (in != null) {
			try {
				try {
					return new NameEnvironmentAnswer(ClassFileReader.read(in, rn, true), null);
				} finally {
					in.close();
				}
			} catch (Exception e) {
				if (e instanceof ClassFormatException) {
					int ec = ((ClassFormatException)e).getErrorCode();
					throw new RuntimeException("Failed to read class file (error "+ec+"): " + rn, e);
				}
				throw new RuntimeException("Failed to read class file " + rn, e);
			}
		}
		return null;
	}

//	private final static Logger logger = Logger.getLogger(NameEnvironmentImpl.class.getName());
}
