/*
 * z2-Environment
 * 
 * Copyright(c) ZFabrik Software GmbH & Co. KG
 * 
 * contact@zfabrik.de
 * 
 * http://www.z2-environment.eu
 */
package com.zfabrik.impl.gitcr.helper;

import java.io.File;
import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.eclipse.jgit.lib.Constants;
import org.eclipse.jgit.lib.Repository;
import org.eclipse.jgit.storage.file.FileRepositoryBuilder;
import org.eclipse.jgit.transport.CredentialsProvider;
import org.eclipse.jgit.transport.RefSpec;
import org.eclipse.jgit.transport.URIish;

import com.zfabrik.impl.gitcr.GitComponentRepositoryImpl;
import com.zfabrik.util.fs.FileUtils;
import com.zfabrik.util.runtime.Foundation;

/**
 * The actual GitCR operations in complete for better testability
 */
public class GitCommand {
	private final static Logger logger = Logger.getLogger(GitComponentRepositoryImpl.class.getName());

	
	private File localRepo;
	private URIish originUri;
	private boolean isOptional;
	private RefSpec fetchRefSpec;
	private CredentialsProvider credentials;
	private int timeout;
	private String toString;
	
	public GitCommand(String toString, URIish originUri, File localRepo, RefSpec fetchRefSpec, boolean isOptional, CredentialsProvider credentials, int timeout) {
		super();
		this.toString = toString;
		this.originUri = originUri;
		this.localRepo = localRepo;
		this.fetchRefSpec = fetchRefSpec;
		this.isOptional = isOptional;
		this.credentials = credentials;
		this.timeout = timeout;
	}

	/*
	 * clone the repository and clean-up on failure
	 */
	public void doClone() {
		try {
			logger.info("Cloning "+this.toString()+" using "+this.fetchRefSpec+" from " + this.originUri);
			// purging the folder allows to call clone in any situation
			FileUtils.delete(this.localRepo);
			// clone the origin repository into the z2 work folder
			try (Repository clonedRepo = GitTools.cloneRepository(
					this.originUri, 
					this.localRepo, 
					this.fetchRefSpec, 
					this.credentials, 
					this.timeout))
			{
				// check out FETCH_HEAD
				String sha = GitTools.checkOutRef(clonedRepo,Constants.FETCH_HEAD);
				logger.info("After cloning "+this.fetchRefSpec.getDestination()+" check out is at "+sha);
			}
		} catch (Exception e) {
			// delete any repository ruins
			FileUtils.delete(this.localRepo);
			logger.log(Level.WARNING, "GIT-CR: Failed cloning " + this.toString(),e);
			// something went wrong - maybe we are offline.
			if (this.isOptional) {
				logger.info("GIT-CR: '" + this.originUri.getHumanishName() + "' will be ignored because it's an optional repository or running in relaxed mode.");

			} else {
				throw new CloneFailedException(toString(),this.localRepo.getAbsolutePath(), e);
			}
		}
	}

	/**
	 * fetch the deltas and try clone on failure
	 */
	public void doFetch() {
		if (Foundation.isOfflineMode()) {
			logger.info("GIT-CR: '" + this.originUri.getHumanishName() + "' will be ignored (Running offline)");
			return;
		}
		logger.info("Fetching "+this.toString()+" using "+this.fetchRefSpec+" from " + this.originUri);
		try (Repository repo = open()) {
			// fetch
			GitTools.fetchRepository(repo, this.fetchRefSpec, credentials, timeout);
			// update working copy
			String sha = GitTools.checkOutRef(repo,Constants.FETCH_HEAD);
			logger.info("After fetching "+this.fetchRefSpec.getDestination()+" check out is at "+sha);
		} catch (Exception e) {

			// fetching failed!

			// just log a short description
			logger.log(Level.WARNING, "GIT-CR: Updating " + this.toString() + " failed with exception",e);

			if (this.isOptional) {
				// origin is not reachable, but repository is optional (maybe we are offline). Log this incident and try to continue
				logger.info("GIT-CR: '" + this.originUri.getHumanishName() + "' will be ignored because it's an optional repository or running in relaxed mode");

			} else {
				// throw exception in production mode
				throw new FetchFailedException(toString(), e);
			}
		}
	}

	private Repository open() throws IOException {
		FileRepositoryBuilder b = new FileRepositoryBuilder();
		return b.setGitDir(new File(this.localRepo, ".git")).build();
	}
	
	@Override
	public String toString() {
		return "GIT-CR: "+this.toString;
	}

}
