/*
 * z2-Environment
 * 
 * Copyright(c) ZFabrik Software GmbH & Co. KG
 * 
 * contact@zfabrik.de
 * 
 * http://www.z2-environment.eu
 */
package com.zfabrik.impl.db.data;

import java.io.PrintWriter;
import java.lang.reflect.Method;
import java.sql.Connection;
import java.sql.SQLException;
import java.sql.SQLFeatureNotSupportedException;
import java.util.concurrent.Callable;
import java.util.logging.Logger;

import javax.sql.DataSource;

import com.zfabrik.impl.db.data.DataSourceResource.DSWU;
import com.zfabrik.util.threading.ThreadUtil;
import com.zfabrik.work.IWorkResource;
import com.zfabrik.work.WorkUnit;

/**
 * Wraps a nested data source implementation so that we can hook up with the {@link WorkUnit}.
 * 
 * @author hb
 *
 */
class DataSourceWrapper implements DataSource {
	/**
	 * 
	 */
	private final DataSourceResource dataSourceWrapper;
	private DataSource lds;
	private Method getParentLogger;

	public DataSourceWrapper(DataSourceResource dataSourceResource, DataSource cbpds) {
		dataSourceWrapper = dataSourceResource;
		this.lds = cbpds;
		// java 6/7 compatibility: check for new methods.
		this.getParentLogger = DataSourceResource.getMethodIfAvailable(this.lds.getClass(),"getParentLogger");
	}

	public Connection getConnection() throws SQLException {
		/**
		 * See http://bugs.mysql.com/bug.php?id=36565
		 */
		return ThreadUtil.cleanContextExecute(new Callable<Connection>() {
			public Connection call() throws Exception {
				if (DataSourceWrapper.this.dataSourceWrapper.enlistInWorkUnit) {
					WorkUnit wu = WorkUnit.queryCurrent();
					if (wu!=null) { 
						String key = DataSourceResource.class.getName() + ":"+ DataSourceWrapper.this.dataSourceWrapper.name;
						IWorkResource wr = WorkUnit.getCurrent().getResource(key);
						DSWU dswu;
						if (wr != null) {
							if (wr instanceof DSWU) {
								dswu = (DSWU) wr;
							} else {
								throw new IllegalStateException("Found invalid work resource (wrong class) "+ wr + " at " + key + ": "+ DataSourceWrapper.this.dataSourceWrapper.name);
							}
						} else {
							try {
								dswu = dataSourceWrapper.new DSWU(lds.getConnection());
							} catch (Exception ex) {
								ex.printStackTrace();
								throw new IllegalStateException("Failed to initialize work resource: "+ DataSourceWrapper.this.dataSourceWrapper.name, ex);
							}
							try {
								WorkUnit.getCurrent().bindResource(key, dswu);
							} catch (Exception e) {
								e.printStackTrace();
								throw new IllegalStateException("Failed to bind resource to work unit: "+ DataSourceWrapper.this.dataSourceWrapper.name, e);
							}
						}
						Connection c = dswu.getConnection();
						if (c == null)
							throw new IllegalStateException("Found stale work resource (no connection anymore) "+ wr + " at " + key + ": " + DataSourceWrapper.this.dataSourceWrapper.name);
						return c;
					} else {
						DataSourceResource.logger.fine("Data source "+DataSourceWrapper.this.dataSourceWrapper.name+" is configured for enlisting in work unit but no work unit was found. Will return unenlisted connection");
					}
				}
				return lds.getConnection();
			}
		});
	}

	public Connection getConnection(String username, String password) throws SQLException {
		throw new UnsupportedOperationException();
	}

	public boolean isWrapperFor(Class<?> iface) throws SQLException {
		return false;
	}

	public <T> T unwrap(Class<T> iface) throws SQLException {
		return null;
	}

	public int getLoginTimeout() throws SQLException {
		return this.lds.getLoginTimeout();
	}

	public PrintWriter getLogWriter() throws SQLException {
		return this.lds.getLogWriter();
	}

	public void setLoginTimeout(int seconds) throws SQLException {
		this.lds.setLoginTimeout(seconds);
	}

	public void setLogWriter(PrintWriter out) throws SQLException {
		this.lds.setLogWriter(out);
	}

	void _kill() {
		dataSourceWrapper._destroy(lds);
	}
	
	// TODO remove when Java 7 can be made mandatory
	public Logger getParentLogger() throws SQLFeatureNotSupportedException {
		return DataSourceResource.getParentLogger(this.lds, this.getParentLogger);
	}
}